# -*- coding: utf-8 -*-

"""
AWS ECR (Elastic Container Registry) Object-Oriented Interface

This module provides an object-oriented interface for interacting with AWS ECR
(Elastic Container Registry) repositories and images. It offers a higher-level
abstraction over the boto3 ECR client, making it easier to manage ECR resources.
"""

import typing as T
import enum
import hashlib
import dataclasses
from datetime import datetime

import botocore.exceptions

if T.TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_ecr.client import ECRClient


@dataclasses.dataclass
class Base:
    def to_dict(self) -> T.Dict[str, T.Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: T.Dict[str, T.Any]) -> "Base":
        return cls(**data)

    @property
    def hash_key(self) -> str:  # pragma: no cover
        """
        Return a unique hash key for the object.
        """
        raise NotImplementedError

    @property
    def fingerprint(self) -> str:  # pragma: no cover
        """
        Return a unique fingerprint for the object.
        """
        return hashlib.md5(self.hash_key.encode("utf-8")).hexdigest()


class ImageTagMutability(str, enum.Enum):
    MUTABLE = "MUTABLE"
    IMMUTABLE = "IMMUTABLE"


@dataclasses.dataclass
class Repository(Base):
    """
    Represents an ECR repository with its properties and methods for retrieval and listing.

    :param registry_id: The AWS account ID associated with the registry.
    :param repository_name: The name of the repository.
    :param repository_uri: The URI of the repository.
    :param repository_arn: The Amazon Resource Name (ARN) of the repository.
    :param created_at: The date and time when the repository was created.
    :param image_tag_mutability: The tag mutability setting for the repository.
    :param image_scanning_configuration: The image scanning configuration for the repository.
    :param encryption_configuration: The encryption configuration for the repository.
    :param response: The full response from the ECR API call.
    """

    registry_id: str = dataclasses.field()
    repository_name: str = dataclasses.field()
    repository_uri: str = dataclasses.field(default=None)
    repository_arn: str = dataclasses.field(default=None)
    created_aAt: datetime = dataclasses.field(default=None)
    image_tag_mutability: str = dataclasses.field(default=None)
    image_scanning_configuration: dict = dataclasses.field(default=None)
    encryption_configuration: dict = dataclasses.field(default=None)
    response: T.Optional[dict] = dataclasses.field(default=None)

    @classmethod
    def from_describe_repositories_response(
        cls,
        response: T.Dict[str, T.Any],
    ):
        """
        Create a Repository instance from the ECR describe_repositories API response.

        Ref:

        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr/client/describe_repositories.html
        """
        return cls(
            registry_id=response["registryId"],
            repository_name=response["repositoryName"],
            repository_uri=response.get("repositoryUri"),
            repository_arn=response.get("repositoryArn"),
            created_aAt=response.get("createdAt"),
            image_tag_mutability=response.get("imageTagMutability"),
            image_scanning_configuration=response.get("imageScanningConfiguration"),
            encryption_configuration=response.get("encryptionConfiguration"),
            response=response,
        )

    @classmethod
    def get(
        cls,
        ecr_client: "ECRClient",
        repository_name: str,
        registry_id: T.Optional[str] = None,
    ) -> T.Optional["Repository"]:
        """
        Retrieve a single repository by name.

        :param ecr_client: The boto3 ECR client.
        :param repository_name: The name of the repository to retrieve.
        :param registry_id: The AWS account ID associated with the registry.
        """
        kwargs = dict(repositoryNames=[repository_name])
        if registry_id:  # pragma: no cover
            kwargs["registryId"] = registry_id
        try:
            res = ecr_client.describe_repositories(**kwargs)
            return cls.from_describe_repositories_response(res["repositories"][0])
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "RepositoryNotFoundException":
                return None
            else:  # pragma: no cover
                raise e

    @classmethod
    def list(
        cls,
        ecr_client: "ECRClient",
        repository_names: T.Optional[T.List[str]] = None,
        registry_id: T.Optional[str] = None,
    ) -> T.Iterator["Repository"]:
        """
        Based on AWS service quota document https://docs.aws.amazon.com/AmazonECR/latest/userguide/service-quotas.html,
        the maximum number of repositories per account is 10,000 per region.

        :param ecr_client: The boto3 ECR client.
        :param repository_names: List of repository names to filter by.
        :param registry_id: The AWS account ID associated with the registry.
        """
        paginator = ecr_client.get_paginator("describe_repositories")
        kwargs = dict(
            PaginationConfig={
                "MaxItems": 10000,
                "PageSize": 1000,
            }
        )
        if registry_id:  # pragma: no cover
            kwargs["registryId"] = registry_id
        if repository_names:
            kwargs["repositoryNames"] = repository_names
        response_iterator = paginator.paginate(**kwargs)
        for response in response_iterator:
            for repo in response.get("repositories", []):
                yield cls.from_describe_repositories_response(repo)


@dataclasses.dataclass
class Image(Base):
    """
    Represents an ECR image with its properties and methods for retrieval and listing.

    :param registry_id: The AWS account ID associated with the registry.
    :param repository_name: The name of the repository containing the image.
    :param image_digest: The sha256 digest of the image manifest.
    :param image_tagsof str): The tags associated with the image.
    :param image_size_in_bytes: The size of the image in bytes.
    :param image_pushed_at: The date and time when the image was pushed.
    :param image_scan_status: The status of the image scan.
    :param image_scan_findings_summary: A summary of findings from the image scan.
    :param image_manifest_media_type: The media type of the image manifest.
    :param artifact_media_type: The media type of the image artifact.
    :param last_recorded_pull_time: The last time the image was pulled.
    :param response: The full response from the ECR API call.
    """

    registry_id: str = dataclasses.field()
    repository_name: str = dataclasses.field()
    image_digest: str = dataclasses.field()
    image_tags: T.List[str] = dataclasses.field(default=None)
    image_size_in_bytes: int = dataclasses.field(default=None)
    image_pushed_at: datetime = dataclasses.field(default=None)
    image_scan_status: dict = dataclasses.field(default=None)
    image_scan_findings_summary: dict = dataclasses.field(default=None)
    image_manifest_media_type: str = dataclasses.field(default=None)
    artifact_media_type: str = dataclasses.field(default=None)
    last_recorded_pull_time: datetime = dataclasses.field(default=None)
    response: T.Optional[dict] = dataclasses.field(default=None)

    @classmethod
    def from_describe_images_response(
        cls,
        response: T.Dict[str, T.Any],
    ):  # pragma: no cover
        """
        Create an Image instance from the ECR describe_images API response.

        Ref:

        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr/client/describe_images.html
        """
        return cls(
            registry_id=response["registryId"],
            repository_name=response["repositoryName"],
            image_digest=response["imageDigest"],
            image_tags=response.get("imageTags"),
            image_size_in_bytes=response.get("imageSizeInBytes"),
            image_pushed_at=response.get("imagePushedAt"),
            image_scan_status=response.get("imageScanStatus"),
            image_scan_findings_summary=response.get("imageScanFindingsSummary"),
            image_manifest_media_type=response.get("imageManifestMediaType"),
            artifact_media_type=response.get("artifactMediaType"),
            last_recorded_pull_time=response.get("lastRecordedPullTime"),
            response=response,
        )

    @classmethod
    def get_by_digest(
        cls,
        ecr_client: "ECRClient",
        repository_name: str,
        image_digest: str,
        registry_id: T.Optional[str] = None,
    ) -> T.Optional["Image"]:  # pragma: no cover
        """
        Retrieve a single image by its digest.

        :param ecr_client: The boto3 ECR client.
        :param repository_name: The name of the repository containing the image.
        :param image_digest: The sha256 digest of the image manifest.
        :param registry_id: The AWS account ID associated with the registry.
        """
        if image_digest.startswith("sha256:") is False:
            image_digest = f"sha256:{image_digest}"
        kwargs = dict(
            repositoryName=repository_name, imageIds=[{"imageDigest": image_digest}]
        )
        if registry_id:
            kwargs["registryId"] = registry_id
        try:
            res = ecr_client.describe_images(**kwargs)
            return cls.from_describe_images_response(res["imageDetails"][0])
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ImageNotFoundException":
                return None
            else:  # pragma: no cover
                raise e

    @classmethod
    def get_by_tag(
        cls,
        ecr_client: "ECRClient",
        repository_name: str,
        image_tag: str,
        registry_id: T.Optional[str] = None,
    ) -> T.Optional["Image"]:  # pragma: no cover
        """
        Retrieve a single image by its tag.

        :param ecr_client: The boto3 ECR client.
        :param repository_name: The name of the repository containing the image.
        :param image_tag: The tag of the image to retrieve.
        :param registry_id: The AWS account ID associated with the registry.
        """
        kwargs = dict(
            repositoryName=repository_name,
            imageIds=[{"imageTag": image_tag}],
        )
        if registry_id:
            kwargs["registryId"] = registry_id
        try:
            res = ecr_client.describe_images(**kwargs)
            return cls.from_describe_images_response(res["imageDetails"][0])
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ImageNotFoundException":
                return None
            else:  # pragma: no cover
                raise e

    @classmethod
    def list(
        cls,
        ecr_client: "ECRClient",
        repository_name: str,
        registry_id: T.Optional[str] = None,
        filter: T.Optional[T.Dict[str, T.Any]] = None,
    ) -> T.Iterator["Image"]:  # pragma: no cover
        """
        List images in an ECR repository.

        :param ecr_client: The boto3 ECR client.
        :param repository_name: The name of the repository containing the image.
        :param registry_id: The AWS account ID associated with the registry.
        :param filter: A dictionary of filters to apply to the list operation.
        """
        paginator = ecr_client.get_paginator("describe_images")
        kwargs = dict(
            repositoryName=repository_name,
            PaginationConfig={
                "MaxItems": 10000,
                "PageSize": 1000,
            },
        )
        if registry_id:
            kwargs["registryId"] = registry_id
        if filter:
            kwargs["filter"] = filter
        response_iterator = paginator.paginate(**kwargs)
        for response in response_iterator:
            for image in response.get("imageDetails", []):
                yield cls.from_describe_images_response(image)


@dataclasses.dataclass
class Destination(Base):
    """
    todo: add docstring
    """

    region: str = dataclasses.field()
    registryId: str = dataclasses.field()

    @property
    def hash_key(self) -> str:
        return f"registryId={self.registryId},region={self.region}"


@dataclasses.dataclass
class RepositoryFilter(Base):
    """
    todo: add docstring
    """

    filter: str = dataclasses.field()
    filterType: str = dataclasses.field(default="PREFIX_MATCH")

    @property
    def hash_key(self) -> str:
        return f"filterType={self.filterType},filter={self.filter}"


@dataclasses.dataclass
class ReplicationRule(Base):
    """
    Represents an ECR replication rule with its properties and methods for retrieval and listing.
    """

    destinations: T.List[Destination] = dataclasses.field()
    repositoryFilters: T.List[RepositoryFilter] = dataclasses.field()

    @classmethod
    def from_dict(cls, data: T.Dict[str, T.Any]):
        destinations = [Destination(**d) for d in data["destinations"]]
        repositoryFilters = [RepositoryFilter(**r) for r in data["repositoryFilters"]]
        return cls(destinations=destinations, repositoryFilters=repositoryFilters)

    @property
    def hash_key(self) -> str:
        destinations = [des.hash_key for des in self.destinations]
        destinations.sort()
        repositoryFilters = [rep.hash_key for rep in self.repositoryFilters]
        repositoryFilters.sort()
        return "{}|{}".format(
            "&".join(destinations),
            "&".join(repositoryFilters),
        )



# -*- coding: utf-8 -*-

from simple_aws_ecr.model import ReplicationRule, Destination, RepositoryFilter
from simple_aws_ecr.recipe import (
    configure_cross_account_lambda_get,
    configure_replication_for_source_registry,
    configure_replication_for_destination_registry,
)
from simple_aws_ecr.tests.mock_aws import BaseMockAwsTest


class Test(BaseMockAwsTest):
    use_mock = True

    def test_replication(self):
        account_devops = self.bsm.aws_account_id
        account_user = "999988887777"

        configure_replication_for_source_registry(
            ecr_client=self.ecr_client,
            rules=[
                ReplicationRule(
                    destinations=[
                        Destination(
                            region=self.bsm.aws_region,
                            registryId=account_user,
                        )
                    ],
                    repositoryFilters=[
                        RepositoryFilter(
                            filter="PREFIX_MATCH",
                            filterType="test",
                        ),
                    ],
                ),
            ],
        )
        configure_replication_for_source_registry(
            ecr_client=self.ecr_client,
            rules=[
                ReplicationRule(
                    destinations=[
                        Destination(
                            region=self.bsm.aws_region,
                            registryId=account_user,
                        )
                    ],
                    repositoryFilters=[
                        RepositoryFilter(
                            filter="PREFIX_MATCH",
                            filterType="test",
                        ),
                    ],
                ),
            ],
        )

        configure_replication_for_destination_registry(
            ecr_client=self.ecr_client,
            source_account_id_list=[account_devops],
            target_account_id=account_user,
            target_region=self.bsm.aws_region,
        )
        configure_replication_for_destination_registry(
            ecr_client=self.ecr_client,
            source_account_id_list=[account_devops],
            target_account_id=account_user,
            target_region=self.bsm.aws_region,
        )

    def test_lambda_get(self):
        account_devops = self.bsm.aws_account_id
        account_user = "999988887777"
        repo_name = "test-repo"
        lbd_func_name_prefix = "test-lbd"

        self.ecr_client.create_repository(repositoryName=repo_name)
        configure_cross_account_lambda_get(
            ecr_client=self.ecr_client,
            repo_name=repo_name,
            aws_account_id_list=[account_user],
            aws_region=self.bsm.aws_region,
            lbd_func_name_prefix=lbd_func_name_prefix,
        )
        configure_cross_account_lambda_get(
            ecr_client=self.ecr_client,
            repo_name=repo_name,
            aws_account_id_list=[account_user],
            aws_region=self.bsm.aws_region,
            lbd_func_name_prefix=lbd_func_name_prefix,
        )


if __name__ == "__main__":
    from aws_ecr.tests import run_cov_test

    run_cov_test(__file__, "simple_aws_ecr.recipe", preview=False)



# -*- coding: utf-8 -*-

"""
A Toolkit for AWS Elastic Container Registry Operations

This module provides a collection of utility functions and classes for common
operations on AWS Elastic Container Registry (ECR). It serves as a "recipe book"
for ECR-related tasks, offering pre-built solutions for frequent use cases.
"""

import typing as T
import json
import hashlib
import dataclasses
from datetime import timezone

import botocore.exceptions

from .model import Image, ReplicationRule
from .utils import get_utc_now

if T.TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_ecr.client import ECRClient


# ------------------------------------------------------------------------------
# Cross Account
# ------------------------------------------------------------------------------
SID_ALLOW_CROSS_ACCOUNT_GET = "AllowCrossAccountGet"
SID_ALLOW_CROSS_ACCOUNT_LBD_GET = "AllowCrossAccountLambdaGet"


@dataclasses.dataclass
class Policy:
    """
    Represents and manages an ECR registry or repository permission policy.

    This class provides methods to create, modify, and convert ECR repository
    permission policies.

    :param document: The full policy document.
    :param statements: A dictionary of policy statements, keyed by their Sid.
    """

    # fmt: off
    document: T.Dict[str, T.Dict[str, T.Any]] = dataclasses.field()
    statements: T.Dict[str, T.Dict[str, T.Any]] = dataclasses.field(default_factory=dict)
    # fmt: on

    @classmethod
    def from_policy_document(
        cls,
        policy_document: T.Dict[str, T.Any],
    ):
        """
        Create a RepositoryPermissionPolicy instance from a policy document.

        :param policy_document: The policy document as a dictionary from the
            ``get_repository_policy`` API.
        """
        return cls(
            document=policy_document,
            statements={
                dct["Sid"]: dct for dct in policy_document.get("Statement", [])
            },
        )

    def to_policy_document(self) -> T.Dict[str, T.Any]:
        """
        Convert the RepositoryPermissionPolicy to a policy document dictionary.
        """
        new_document = json.loads(json.dumps(self.document))
        new_document["Statement"] = list(
            sorted(
                self.statements.values(),
                key=lambda dct: dct["Sid"],
            )
        )
        return new_document


def build_cross_account_get_statement(
    aws_account_id_list: T.List[str],
) -> T.Dict[str, T.Any]:
    """
    Build a policy statement allowing cross-account access to ECR images.

    :param aws_account_id_list: List of AWS account IDs to grant access.
    """
    return {
        "Sid": SID_ALLOW_CROSS_ACCOUNT_GET,
        "Effect": "Allow",
        "Principal": {
            "AWS": [
                f"arn:aws:iam::{aws_account_id}:root"
                for aws_account_id in aws_account_id_list
            ]
        },
        "Action": [
            "ecr:BatchGetImage",
            "ecr:GetDownloadUrlForLayer",
        ],
    }


def build_cross_account_lbd_get_statement(
    aws_account_id_list: T.List[str],
    aws_region: str = "*",
    lbd_func_name_prefix: str = "",
):
    """
    Build a policy statement allowing cross-account Lambda function access to ECR images.

    .. note::

        You also need to add the permission defined in
        :func:`build_cross_account_get_statement`

    :param aws_account_id_list: List of AWS account IDs to grant access.
    :param aws_region: AWS region for Lambda functions. Defaults to "*" (all regions).
    :param lbd_func_name_prefix: Prefix for allowed Lambda function names.
        Defaults to "" (all Lambda functions).
    """
    return {
        "Sid": SID_ALLOW_CROSS_ACCOUNT_LBD_GET,
        "Effect": "Allow",
        "Principal": {"Service": "lambda.amazonaws.com"},
        "Action": [
            "ecr:BatchGetImage",
            "ecr:GetDownloadUrlForLayer",
        ],
        "Condition": {
            "StringLike": {
                "aws:sourceARN": [
                    f"arn:aws:lambda:{aws_region}:{aws_account_id}:function:{lbd_func_name_prefix}*"
                    for aws_account_id in aws_account_id_list
                ]
            }
        },
    }


def configure_cross_account_lambda_get(
    ecr_client: "ECRClient",
    repo_name: str,
    aws_account_id_list: T.List[str],
    aws_region: str = "*",
    lbd_func_name_prefix: str = "",
):
    """
    Configure cross-account access for Lambda functions to an ECR repository.

    This function updates the repository policy to allow specified AWS accounts
    and their Lambda functions to pull images from the repository.

    :param ecr_client: The boto3 ECR client.
    :param repo_name: Name of the ECR repository.
    :param aws_account_id_liststr: List of AWS account IDs to grant access.
    :param aws_region: AWS region for Lambda functions. Defaults to "*" (all regions).
    :param lbd_func_name_prefix: Prefix for allowed Lambda function names.
        Defaults to "" (all Lambda functions).
    """
    # get existing
    try:
        response = ecr_client.get_repository_policy(
            repositoryName=repo_name,
        )
        policy_document = json.loads(response["policyText"])
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "RepositoryPolicyNotFoundException":
            policy_document = {
                "Version": "2012-10-17",
                "Statement": [],
            }
        else:  # pragma: no cover
            raise e

    policy = Policy.from_policy_document(policy_document)
    is_policy_changed = False

    # acc statement
    # NOTE: even though it sounds like we only need the lambda function service
    # as principal, but we still need the root account as principal to allow
    # the lambda function to pull images from the ECR repository
    acc_stmt = policy.statements.get(
        SID_ALLOW_CROSS_ACCOUNT_GET,
        build_cross_account_get_statement(
            aws_account_id_list=[],
        ),
    )
    arn_list: T.List[str] = acc_stmt["Principal"]["AWS"]
    for aws_account_id in aws_account_id_list:
        arn = f"arn:aws:iam::{aws_account_id}:root"
        if arn not in arn_list:
            arn_list.append(arn)
            is_policy_changed = True
    if aws_account_id_list:
        policy.statements[SID_ALLOW_CROSS_ACCOUNT_GET] = acc_stmt

    # lbd statement
    lbd_stmt = policy.statements.get(
        SID_ALLOW_CROSS_ACCOUNT_LBD_GET,
        build_cross_account_lbd_get_statement(
            aws_account_id_list=[],
        ),
    )
    arn_list: T.List[str] = lbd_stmt["Condition"]["StringLike"]["aws:sourceARN"]
    for aws_account_id in aws_account_id_list:
        arn = f"arn:aws:lambda:{aws_region}:{aws_account_id}:function:{lbd_func_name_prefix}*"
        if arn not in arn_list:
            arn_list.append(arn)
            is_policy_changed = True
    if aws_account_id_list:
        policy.statements[SID_ALLOW_CROSS_ACCOUNT_LBD_GET] = lbd_stmt

    # make change
    if is_policy_changed:
        ecr_client.set_repository_policy(
            repositoryName=repo_name,
            policyText=json.dumps(policy.to_policy_document(), indent=4),
        )


# ------------------------------------------------------------------------------
# Replication
# ------------------------------------------------------------------------------
def configure_replication_for_source_registry(
    ecr_client: "ECRClient",
    rules: T.List["ReplicationRule"],
):
    """
    Configure ECR replication rules for a source registry.

    This function retrieves existing replication rules, merges them with the provided
    rules, and updates the registry's replication configuration.

    .. note::

        This function retrieves existing replication rules, merges them with the provided
        rules, and updates the registry's replication configuration. It is an idempotent
        operation that adds new rules and updates existing ones without removing any
        rules not present in the input.

    :param ecr_client: The boto3 ECR client.
    :param rules: A list of :class:`aws_ecr.model.ReplicationRule` objects to be applied.

    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr/client/put_replication_configuration.html
    """
    res = ecr_client.describe_registry()
    existing_rule_list = [
        ReplicationRule.from_dict(dct)
        for dct in res.get("replicationConfiguration", {}).get("rules", [])
    ]
    existing_rule_dict = {rule.fingerprint: rule for rule in existing_rule_list}

    for rule in rules:
        existing_rule_dict[rule.fingerprint] = rule

    ecr_client.put_replication_configuration(
        replicationConfiguration={
            "rules": [rule.to_dict() for rule in existing_rule_dict.values()],
        },
    )


def configure_replication_for_destination_registry(
    ecr_client: "ECRClient",
    source_account_id_list: T.List[str],
    target_account_id: str,
    target_region: str,
):
    """
    Configure the registry policy for an ECR destination registry to allow replication.

    This function creates or updates the registry policy to permit specified source
    accounts to replicate images to the destination registry.

    :param ecr_client: The boto3 ECR client for the destination registry.
    :param source_account_id_list: List of AWS account IDs allowed to replicate to this registry.
    :param target_account_id: The AWS account ID of the destination registry.
    :param target_region: The AWS region of the destination registry.

    Ref:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr/client/get_registry_policy.html
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr/client/put_registry_policy.html
    """
    # get existing
    try:
        response = ecr_client.get_registry_policy()
        policy_document = json.loads(response["policyText"])
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "RegistryPolicyNotFoundException":
            policy_document = {
                "Version": "2012-10-17",
                "Statement": [],
            }
        else:  # pragma: no cover
            raise e

    policy = Policy.from_policy_document(policy_document)

    source_account_id_list.sort()
    key = "{}-{}-{}".format(
        ",".join(source_account_id_list),
        target_account_id,
        target_region,
    )
    sid = "Replication{}".format(hashlib.md5(key.encode("utf8")).hexdigest())
    statement = {
        "Sid": sid,
        "Effect": "Allow",
        "Principal": {
            "AWS": [
                f"arn:aws:iam::{account_id}:root"
                for account_id in source_account_id_list
            ],
        },
        "Action": [
            "ecr:CreateRepository",
            "ecr:ReplicateImage",
        ],
        "Resource": [f"arn:aws:ecr:{target_region}:{target_account_id}:repository/*"],
    }
    policy.statements[sid] = statement
    ecr_client.put_registry_policy(
        policyText=json.dumps(policy.to_policy_document(), indent=4),
    )


def delete_untagged_image(
    ecr_client: "ECRClient",
    repo_name: str,
    expire: int = 90 * 24 * 60 * 60,
) -> T.List[Image]: # pragma: no cover
    """
    Delete untagged images from an ECR repository.

    This function lists all untagged images in the repository and deletes them.

    :param ecr_client: The boto3 ECR client.
    :param repo_name: Name of the ECR repository.
    :param expire: The number of seconds after which untagged images are deleted.
    """
    to_delete_image_list = list()
    utc_now = get_utc_now()
    for image in Image.list(
        ecr_client=ecr_client,
        repository_name=repo_name,
        filter={"tagStatus": "UNTAGGED"},
    ):
        image_pushed_at = image.image_pushed_at
        if image_pushed_at.tzinfo is None:
            image_pushed_at = image_pushed_at.replace(tzinfo=timezone.utc)
        elapse = (image_pushed_at - utc_now).total_seconds()
        if elapse > expire:
            to_delete_image_list.append(image)
        if to_delete_image_list:
            ecr_client.batch_delete_image(
                repositoryName=repo_name,
                imageIds=[
                    {"imageDigest": img.image_digest} for img in to_delete_image_list
                ],
            )
        return to_delete_image_list



# -*- coding: utf-8 -*-

"""
Make docker cli easier to work with AWS ECR.
"""

import typing as T
import base64
import subprocess
import dataclasses
from pathlib import Path

from .vendor.better_pathlib import temp_cwd

if T.TYPE_CHECKING:  # pragma: no cover
    from boto_session_manager import BotoSesManager
    from mypy_boto3_ecr.client import ECRClient


def get_ecr_registry_url(
    aws_account_id: str,
    aws_region: str,
):  # pragma: no cover
    """
    Get the full ECR registry URL.

    :param aws_account_id: The AWS account id of your ECR repo.
    :param aws_region: The AWS region of your ECR repo
    """
    return f"https://{aws_account_id}.dkr.ecr.{aws_region}.amazonaws.com"


def get_ecr_image_uri(
    aws_account_id: str,
    aws_region: str,
    ecr_repo_name: str,
    tag: str,
) -> str:  # pragma: no cover
    """
    Get the full ECR repo URI with image tag.

    :param aws_account_id: The AWS account id of your ECR repo.
    :param aws_region: The AWS region of your ECR repo
    :param ecr_repo_name: the ECR repo name
    :param tag: the image tag
    """
    return f"{aws_account_id}.dkr.ecr.{aws_region}.amazonaws.com/{ecr_repo_name}:{tag}"


def get_ecr_auth_token(
    ecr_client: "ECRClient",
) -> T.Tuple[str, str]:  # pragma: no cover
    """
    Get ECR auth token using boto3 SDK.

    Reference:

    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr/client/get_authorization_token.html

    :return: username and password
    """
    res = ecr_client.get_authorization_token()
    b64_token = res["authorizationData"][0]["authorizationToken"]
    user_pass = base64.b64decode(b64_token.encode("utf-8")).decode("utf-8")
    username, password = user_pass.split(":", 1)
    return username, password


def docker_login(
    username: str,
    password: str,
    registry_url: str,
) -> bool:  # pragma: no cover
    """
    Login docker cli to AWS ECR. Run:

    .. code-block:: bash

        echo ${password} | docker login -u ${username} --password-stdin ${registry_url}

    Reference:

    - https://docs.aws.amazon.com/AmazonECR/latest/userguide/registry_auth.html#registry-auth-token

    :return: a boolean flag to indicate if the login is successful.
    """
    args = ["echo", password]
    pipe = subprocess.Popen(args, stdout=subprocess.PIPE)
    args = ["docker", "login", "-u", username, registry_url, "--password-stdin"]
    response = subprocess.run(args, stdin=pipe.stdout, capture_output=True)
    text = response.stdout.decode("utf-8")
    is_succeeded = "Login Succeeded" in text
    return is_succeeded


def ecr_login(
    ecr_client: "ECRClient",
    aws_account_id: str,
    aws_region: str,
) -> bool:  # pragma: no cover
    """
    Login docker cli to AWS ECR using boto3 SDK and AWS CLI.

    :return: a boolean flag to indicate if the login is successful.
    """
    registry_url = get_ecr_registry_url(
        aws_account_id=aws_account_id,
        aws_region=aws_region,
    )
    username, password = get_ecr_auth_token(ecr_client=ecr_client)
    return docker_login(username, password, registry_url)


@dataclasses.dataclass
class EcrContext:
    """
    A utility class to help build and push docker image to ECR.

    :param aws_account_id: The AWS account id of your ECR repo.
    :param aws_region: The AWS region of your ECR repo
    :param repo_name: the repo name
    :param dir_dockerfile: the directory where the Dockerfile is located.
    """

    aws_account_id: str = dataclasses.field()
    aws_region: str = dataclasses.field()
    repo_name: str = dataclasses.field()
    path_dockerfile: Path = dataclasses.field()

    @classmethod
    def from_bsm(
        cls,
        bsm: "BotoSesManager",
        repo_name: str,
        path_dockerfile: Path,
    ):  # pragma: no cover
        """
        Create a new instance of EcrContext.

        :param bsm: ``boto_session_manager.BotoSesManager`` object.
        :param repo_name: ECR repo name.
        :param path_dockerfile: the path to the Dockerfile.
        """
        return cls(
            aws_account_id=bsm.aws_account_id,
            aws_region=bsm.aws_region,
            repo_name=repo_name,
            path_dockerfile=path_dockerfile,
        )

    @property
    def dir_dockerfile(self) -> Path:  # pragma: no cover
        return self.path_dockerfile.parent

    def get_image_uri(self, tag: str) -> str:  # pragma: no cover
        """
        Get the ECR container image URI.

        :param tag: the container image tag.
        """
        return get_ecr_image_uri(
            aws_account_id=self.aws_account_id,
            aws_region=self.aws_region,
            ecr_repo_name=self.repo_name,
            tag=tag,
        )

    def build_image(
        self,
        image_tag_list: T.Optional[T.List[str]] = None,
        additional_args: T.Optional[T.List[str]] = None,
    ):  # pragma: no cover
        """
        Build docker image.

        :param image_tag_list: the list of tag you want to give to the built image,
        e.g. ["latest", "0.1.1"]
        :param additional_args: additional command line arguments for ``docker build ...``

        .. note::

            If you are trying to build a linux/amd64 compatible image on an ARM chip Mac
            you need to set ``"--platform=linux/amd64"`` in the ``additional_args``.
        """
        if image_tag_list is None:
            image_tag_list = ["latest"]
        if additional_args is None:
            additional_args = []
        with temp_cwd(self.dir_dockerfile):
            args = ["docker", "build"]
            # Reference: https://docs.docker.com/engine/reference/commandline/build/#tag
            for tag in image_tag_list:
                args.extend(["-t", self.get_image_uri(tag)])
            args.extend(additional_args)
            args.append(".")
            # don't use check=True
            # the args may include sensitive information like aws credentials
            # we don't want to automatically print to the log
            # instead, we want to handle the error our self.
            result = subprocess.run(args)
            if result.returncode != 0:
                raise subprocess.CalledProcessError(
                    result.returncode,
                    "'docker build' command did not exit successfully!",
                )

    def push_image(
        self,
        image_tag_list: T.Optional[T.List[str]] = None,
        additional_args: T.Optional[T.List[str]] = None,
    ):  # pragma: no cover
        """
        Push Docker image to ECR.
        """
        if image_tag_list is None:
            image_tag_list = ["latest"]
        if additional_args is None:
            additional_args = []
        with temp_cwd(self.dir_dockerfile):
            for tag in image_tag_list:
                args = [
                    "docker",
                    "push",
                    self.get_image_uri(tag),
                ]
                args.extend(additional_args)
                subprocess.run(args, check=True)

    def test_image(
        self,
        tag: T.Optional[str] = None,
        additional_args: T.Optional[T.List[str]] = None,
    ):  # pragma: no cover
        """
        Test the container image locally by running it.

        :param additional_args: additional command line arguments for ``docker run ...``
        """
        if tag is None:
            tag = "latest"
        if additional_args is None:
            additional_args = []
        with temp_cwd(self.dir_dockerfile):
            image_uri = self.get_image_uri(tag)
            args = [
                "docker",
                "run",
                "--rm",
                image_uri,
            ]
            args.extend(additional_args)
            subprocess.run(args, check=True)



# -*- coding: utf-8 -*-

from simple_aws_ecr import api


def test():
    _ = api
    _ = api.get_ecr_registry_url
    _ = api.get_ecr_image_uri
    _ = api.get_ecr_auth_token
    _ = api.docker_login
    _ = api.ecr_login
    _ = api.EcrContext
    _ = api.Repository
    _ = api.Image
    _ = api.Destination
    _ = api.RepositoryFilter
    _ = api.ReplicationRule
    _ = api.configure_replication_for_source_registry
    _ = api.configure_replication_for_destination_registry
    _ = api.SID_ALLOW_CROSS_ACCOUNT_GET
    _ = api.SID_ALLOW_CROSS_ACCOUNT_LBD_GET
    _ = api.configure_cross_account_lambda_get
    _ = api.delete_untagged_image


if __name__ == "__main__":
    from aws_ecr.tests import run_cov_test

    run_cov_test(__file__, "simple_aws_ecr.api", preview=False)



# -*- coding: utf-8 -*-

from simple_aws_ecr.model import Repository
from simple_aws_ecr.tests.mock_aws import BaseMockAwsTest


class Test(BaseMockAwsTest):
    def test(self):
        repo_name = "test-repo"
        repo = Repository.get(
            ecr_client=self.ecr_client,
            repository_name=repo_name,
        )
        assert repo is None

        self.ecr_client.create_repository(repositoryName=repo_name)

        repo = Repository.get(
            ecr_client=self.ecr_client,
            repository_name=repo_name,
        )

        repo_list = list(Repository.list(ecr_client=self.ecr_client))
        assert len(repo_list) == 1
        assert repo_list[0].repository_name == repo_name


if __name__ == "__main__":
    from aws_ecr.tests import run_cov_test

    run_cov_test(__file__, "simple_aws_ecr.model", preview=False)



# -*- coding: utf-8 -*-

from .docker import get_ecr_registry_url
from .docker import get_ecr_image_uri
from .docker import get_ecr_auth_token
from .docker import docker_login
from .docker import ecr_login
from .docker import EcrContext
from .model import Repository
from .model import Image
from .model import Destination
from .model import RepositoryFilter
from .model import ReplicationRule
from .recipe import configure_replication_for_source_registry
from .recipe import configure_replication_for_destination_registry
from .recipe import SID_ALLOW_CROSS_ACCOUNT_GET
from .recipe import SID_ALLOW_CROSS_ACCOUNT_LBD_GET
from .recipe import configure_cross_account_lambda_get
from .recipe import delete_untagged_image
